"""
![test](https://github.com/PyMLTools/imageapply/actions/workflows/tests.yml/badge.svg)

A package for improving the simplicity of adapting Segmentation Models to large images.

This package include a class that can be used to apply a segmentation model to large images. The class will split the image into smaller tiles, apply the model to each tile, and then stitch the tiles back together. This allows the model to be applied to images that are too large to fit in memory.

## Installation

```bash
pip install imageapply
```

## Usage
This package can be used to easily extend a segmentation model to large images. The following example is for a `model` that takes a batch of images of size (None, 256, 256, 3) as input and outputs a mask of size (None, 256, 256, 1).

```python
from imageapply import FlexibleModel
model = FlexibleModel(model, input_size=(256, 256, 3), max_batch_size=16)

# Apply the model to a batch of large images of size (20, 1024, 1024, 3)
masks = model(images)
``` 
"""


from .flexible_model import *



import numpy as np
from typing import Callable
from .data import T


def apply_model(model : Callable[[T], T], data : T, batch_size=None) -> T:
    """
    Applies the model to a batch of data, running at most batch_size entries through at once.
    
    Args:
        model (Callable[[T], T]): The model to apply
        data (T): The data to apply the model to
        batch_size (int): The maximum number of entries to run through the model at once. If None, \
            runs all entries through at once.
        
    Returns:
        T: The output of the model 
    """

    if batch_size is None or batch_size > data.shape[0]:
        return model(data)

    assert type(batch_size) is int and batch_size > 0, "Batch size must be a positive integer"

    results = []
    for i in range(0, data.shape[0], batch_size):
        results.append(model(data[i:i+min(batch_size, data.shape[0]-i)]))

    return np.concatenate(results, axis=0)

def divide_into_regions(data : T, region_size : tuple) -> T:
    """
    Divides the data into regions of size region_size.
    
    Args:
        data (T): The data to divide
        region_size (tuple): The size of the regions
    
    Returns:
        T: The divided data
    """

    # Assumes data is a multiple of region_size
    # data is array of shape (Batch_size, ...)
    # region_size is tuple of length data.ndim - 1

    n_splits = [data.shape[i+1] // region_size[i] for i in range(len(region_size))]
    for dim in range(len(region_size)):
        data = np.concatenate(np.split(data, n_splits[dim], axis=dim+1), axis=0)
    
    return data

def combine_regions(data : T, region_size : tuple, original_size : tuple) -> T:
    """
    Combines the regions of the data into the original shape.

    Args:
        data (T): The data to combine
        region_size (tuple): The size of the regions
        original_size (tuple): The original shape of the data

    Returns:
        T: The combined data
    """
    n_splits = [original_size[i+1] // region_size[i] for i in range(len(region_size))]
    for dim in range(len(region_size)-1, -1, -1):
        data = np.concatenate(np.split(data, n_splits[dim], axis=0), axis=dim+1)

    return data

####################################################################################################
###################################### PADDING AND CROPPING ########################################
####################################################################################################

def pad_to_multiple(data:T, input_size:tuple, pad_mode:str="zeros", pad_position:str="end") -> T:
    """
    Pads the data to a multiple of the input size.

    Args:
        data (T): The data to pad
        input_size (tuple): The input size of the model
        pad_mode (str): The mode to pad with. One of "zeros", "reflect", "symmetric"
        pad_position (str): The position to pad. One of "end" or "centre"

    Returns:
        T: The combined data
    """

    rem_per_dim = [data.shape[i+1] % input_size[i] for i in range(len(input_size))]
    pad_amount = [input_size[i] - rem_per_dim[i] if rem_per_dim[i] > 0 else 0 for i in range(len(input_size))]
    if pad_position == "end":
        pad_amount = [(0, 0)] + [(0, pad_amount[i]) for i in range(len(pad_amount))]
    elif pad_position == "centre":
        pad_amount = [(0, 0)] + [(pad_amount[i]//2, pad_amount[i] - pad_amount[i]//2) for i in range(len(pad_amount))]
    else:
        raise ValueError("Invalid pad position. Must be 'end' or 'centre'")

    if pad_mode == "zeros":
        return np.pad(data, pad_amount, mode='constant', constant_values=0)
    elif pad_mode == "reflect":
        return np.pad(data, pad_amount, mode='reflect')
    elif pad_mode == "symmetric":
        return np.pad(data, pad_amount, mode='symmetric')
    else:
        raise ValueError("Invalid pad mode. Must be 'zeros', 'reflect', or 'symmetric'")

def crop_to_original(data : T, original_shape : tuple, pad_position : str = "end") -> T:
    """
    Crops the data to the original shape.
    
    Args:
        data (T): The data to crop
        original_shape (tuple): The original shape of the data
        pad_position (str): The position to pad. One of "end" or "centre"
        
    Returns:
        T: The cropped data
    """
    if pad_position == "end":
        crop_indices = tuple([slice(None)] + [slice(0, original_shape[i+1]) for i in range(len(original_shape)-1)])
    elif pad_position == "centre":
        pad_amount = [data.shape[i+1] - original_shape[i+1] for i in range(len(original_shape)-1)]
        crop_indices = tuple([slice(None)] + [slice(pad_amount[i]//2, pad_amount[i]//2 + original_shape[i+1]) for i in range(len(original_shape)-1)])
    return data[crop_indices]


#TODO: Improve Package Name
#TODO: Add support for single image calls (i.e. without batch dimension)
#TODO: Add support for non-batched models (models that can only handle one image at a time)
#TODO: Add support for pytorch tensors (and others)
#TODO: Limit image dims to 3
#TODO: Implement a Data class (with Pytorch, Numpy, and Tensorflow subclasses) to handle the different data types

from .revtransform import PadCrop, DivideCombine, CombinedModel, BasicTTA

class FlexibleModel:
    """
    This model is designed to be used with models that can only handle a certain input size.
    """

    def __init__(self, model, input_size, max_batch_size=None, basic_tta=False):
        """
        Creates a new FlexibleModel object.
        
        Args:
            model (function): The model to apply to the data
            input_size (tuple): The size of the input to the model
            max_batch_size (int): The maximum batch size to use when applying the model
            basic_tta (bool): Whether to use basic test time augmentation
            
        Returns:
            FlexibleModel: The new FlexibleModel object
        """
        assert model is not None and callable(model), "Model must be callable"
        assert input_size[0] is None, "First dimension of input size must be None"

        self.model = model
        self.input_size = input_size[1:]
        self.max_batch_size = max_batch_size
        self.tta = basic_tta
        
        self.combined = CombinedModel([
            BasicTTA() if self.tta else None,
            PadCrop(self.input_size, pad_mode="zeros", pad_position="end"),
            DivideCombine(self.input_size),
            model
        ]) 
    
    def __call__(self, batch):
        """
        Runs the model on the batch of data.
        
        Args:
            batch (T): The batch of data to run the model on
        
        Returns:
            T: The output of the model
        """
        # Assume for now, data is a batch numpy array
        return self.combined(batch) 


from abc import ABC, abstractmethod, abstractproperty
import numpy as np
import torch
from typing import TypeVar

T = TypeVar("T", np.ndarray, torch.Tensor)

class data_object(ABC):

    def __init__(self, data):
        self.data = data

    @abstractproperty
    def shape(self):
        pass

    @abstractmethod
    def copy(self):
        pass

    @abstractmethod
    def split(self, n_splits, axis):
        pass

    @abstractmethod
    def concatenate(self, axis):
        pass

    @abstractmethod
    def __getitem__(self, key):
        pass

class numpy_data(data_object):

    def __init__(self, data):
        super().__init__(data)

    @property
    def shape(self):
        return self.data.shape

    def copy(self):
        return numpy_data(self.data.copy())

    def split(self, n_splits, axis):
        return numpy_data(np.split(self.data, n_splits, axis=axis))

    def concatenate(self, axis):
        return numpy_data(np.concatenate(self.data, axis=axis))

    def __getitem__(self, key):
        return numpy_data(self.data[key])


# Preprocessing tools for training
from .data import data_object

def divide_into_regions(data : data_object, region_size : tuple, pad_mode : str) -> data_object:
    """
    Divides the data into regions of size region_size.
    """

    assert len(region_size) == len(data.shape) - 1, "Region size must have same number of dimensions as data"




from abc import abstractmethod
from typing import Callable, List, Tuple, Union
from .tools import pad_to_multiple, crop_to_original, divide_into_regions, combine_regions
from .data import T
import numpy as np

class CombinedModel:
    """
    A model with all combined transformations. Can be used to sandwich multiple ReversibleTransformations around a base model.
    """
    
    def __init__(self, steps: List[Callable]):
        self.steps = steps
        
        model = self.steps[-1]
        for step in reversed(self.steps[:-1]):
            if step is not None:
                model = step(model)
        self.model = model
        
    def __call__(self, data:T) -> T:
        return self.model(data)
        

class ReversibleTransformation:
    """
    A data transformation that can be reversed.
    """
    
    def __init__(self, distributed:bool=False):
        self.distributed = distributed
        self._inter = id

    def _attach_intermediary(self, intermediary:Callable[[T], T]):
        self._inter = intermediary
    
    @abstractmethod
    def _forward(self, data):
        raise NotImplementedError("forward not implemented")
    
    @abstractmethod
    def _backward(self, data):
        raise NotImplementedError("backward not implemented")
    
    def __call__(self, data_or_inter:Union[T, Callable]) -> T:
        if isinstance(data_or_inter, Callable):
            self._attach_intermediary(data_or_inter)
            return self
        elif self.distributed:
            return self._backward([self._inter(x) for x in self._forward(data_or_inter)])
        else:
            return self._backward(self._inter(self._forward(data_or_inter)))
        
class BasicTTA(ReversibleTransformation):
    """
    Basic Test Time Augmentation. Flips image horizontally, vertically and both.
    """
    
    def __init__(self, combine_mode:str="mean"):
        super().__init__(distributed=True)
        self.combine_mode = combine_mode
    
    def _forward(self, data):
        return [
            data,
            np.flip(data, 1),
            np.flip(data, 2),
            np.flip(data, (1, 2))
        ]
    
    def _backward(self, data):
        data = [
            data[0],
            np.flip(data[1], 1),
            np.flip(data[2], 2),
            np.flip(data[3], (1, 2))
        ]
        
        if self.combine_mode == 'mean':
            return np.mean(data, axis=0)
        else:
            raise ValueError(f"Unknown combine mode {self.combine_mode}")
        
    def __repr__(self) -> str:
        return f"BasicTTA(combine_mode={self.combine_mode})"


class PadCrop(ReversibleTransformation):
    """
    Pad the data to a multiple of the input size, reverse crops the data to the original shape.
    """
    def __init__(self, input_size, pad_mode="reflect", pad_position="end"):
        super().__init__()
        self.input_size = input_size
        self.pad_mode = pad_mode
        self.pad_position = pad_position
    
    def _forward(self, data):
        self.original_shape = data.shape
        return pad_to_multiple(data, self.input_size, pad_mode=self.pad_mode, pad_position=self.pad_position)
    
    def _backward(self, data):
        return crop_to_original(data, self.original_shape, pad_position=self.pad_position)
    
    def __repr__(self):
        return f"PadCrop(input_size={self.input_size}, pad_mode={self.pad_mode}, pad_position={self.pad_position})"
    
class DivideCombine(ReversibleTransformation):
    """
    Divide the data into regions of size region_size, reverse combines the regions into the original shape.
    """
    def __init__(self, region_size:Tuple):
        """
        Args:
            region_size: The size of the regions to divide the data into. 
        """
        super().__init__()
        self.region_size = region_size
    
    def _forward(self, data):
        """
        Forward pass, divide the data into regions.
        """
        self.original_shape = data.shape # save the original shape for the backward pass
        return divide_into_regions(data, self.region_size)
    
    def _backward(self, data):
        return combine_regions(data, self.region_size, self.original_shape)
    
    def __repr__(self):
        return f"DivideCombine(region_size={self.region_size})"


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['imageapply']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1,<2', 'torch>=1,<2']

setup_kwargs = {
    'name': 'imageapply',
    'version': '0.2.1',
    'description': 'Simplifies the process of applying U-net style models to large images.',
    'long_description': '# Image Apply\n![test](https://github.com/PyMLTools/imageapply/actions/workflows/tests.yml/badge.svg)\n\nA package for improving the simplicity of adapting Segmentation Models to large images.\n\nThis package include a class that can be used to apply a segmentation model to large images. The class will split the image into smaller tiles, apply the model to each tile, and then stitch the tiles back together. This allows the model to be applied to images that are too large to fit in memory.\n\n## Installation\n\n```bash\npip install imageapply\n```\n\n## Usage\nThis package can be used to easily extend a segmentation model to large images. The following example is for a `model` that takes a batch of images of size (None, 256, 256, 3) as input and outputs a mask of size (None, 256, 256, 1).\n\n```python\nfrom imageapply import FlexibleModel\nmodel = FlexibleModel(model, input_size=(256, 256, 3), max_batch_size=16)\n\n# Apply the model to a batch of large images of size (20, 1024, 1024, 3)\nmasks = model(images)\n``` \n',
    'author': 'fynnsu',
    'author_email': 'fynnsu@outlook.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
